/*
 * kernel/power/atomic_copy.c
 *
 * Copyright 2004-2006 Nigel Cunningham <nigel@suspend2.net>
 * Copyright (C) 2006 Red Hat, inc.
 *
 * Distributed under GPLv2.
 *
 * Routines for doing the atomic save/restore.
 */

#include <linux/suspend.h>
#include <linux/highmem.h>
#include <linux/cpu.h>
#include <linux/freezer.h>
#include "suspend.h"
#include "storage.h"
#include "power_off.h"
#include "ui.h"
#include "power.h"
#include "io.h"
#include "prepare_image.h"
#include "checksum.h"
#include "suspend_builtin.h"

int extra_pd1_pages_used;

/*
 * Highmem related functions (x86 only).
 */

#ifdef CONFIG_HIGHMEM
static unsigned long origoffset;
static unsigned long copyoffset;
static int o_zone_num, c_zone_num;

struct zone_data {
	unsigned long start_pfn;
	unsigned long end_pfn;
	int is_highmem;
};

/**
 * free_nosave_highmem_info: Free memory allocated for restoring high pages.
 *
 * When aborting a resume prior to doing the atomic restore, free memory
 * allocated for restoring highmem.
 **/

static void free_nosave_highmem_info(void)
{
	if (suspend2_nosave_zone_list)
		free_page((unsigned long) suspend2_nosave_zone_list);
}

/**
 * suspend_init_nosave_highmem_info: Set up nosave data needed for highmem.
 *
 * Zone information might be overwritten during the copy back, so we copy
 * the fields we need to a non-conflicting page and use it, together with
 * pointers to the bitmaps for pageset1. The zone data should be the same
 * in the boot and resumed kernels, but a little paranoia is not necessarily
 * a bad thing.
 **/

static int suspend_init_nosave_highmem_info(void)
{
	struct zone *zone;
	
	suspend2_nosave_origmap = pageset1_map;
	suspend2_nosave_copymap = pageset1_copy_map;
	suspend2_nosave_max_pfn = max_pfn;

	suspend2_nosave_zone_list =
		(struct zone_data *) __suspend_get_nonconflicting_page();

	if (!suspend2_nosave_zone_list)
		return -ENOMEM;

	/* It should all fit in one page. */
	if (ZONEID_MASK > (PAGE_SIZE / sizeof(struct zone_data))) {
		printk("Suspend2: Zone data won't fit in a page.\n");
		return 1;
	}

	memset((char *) suspend2_nosave_zone_list, 0, PAGE_SIZE);

	for_each_zone(zone) {
		int zone_num = page_zone_id(pfn_to_page(zone->zone_start_pfn));
		struct zone_data *this = &suspend2_nosave_zone_list[zone_num];

		if (!populated_zone(zone))
			continue;

		this->start_pfn = zone->zone_start_pfn;
		this->end_pfn = zone->zone_start_pfn + zone->spanned_pages - 1;
		this->is_highmem = is_highmem(zone);
	}

	return 0;
}

/**
 * __suspend_get_next_bit_on: Atomic-copy safe location of next bit on in a bitmap.
 * 
 * @bitmap:	The bitmap we are traversing.
 * @zone_num:	Which zone we are in.
 * @counter:	The current pfn.
 *
 * A version of __get_next_bit_on that can be used when doing the atomic
 * copy. While doing it, we can't rely on the zone information being in
 * a constant location.
 **/

static unsigned long __suspend_get_next_bit_on(dyn_pageflags_t bitmap,
		int *zone_num, unsigned long counter)
{
	unsigned long *ul_ptr = NULL;
	int reset_ul_ptr = 1, pagebit;

	if (counter == (suspend2_nosave_max_pfn + 1)) {
		*zone_num = 0;
		while (!suspend2_nosave_zone_list[*zone_num].end_pfn)
			(*zone_num)++;
		counter = suspend2_nosave_zone_list[*zone_num].start_pfn - 1;
	}

	do {
		counter++;
		if (counter > suspend2_nosave_zone_list[*zone_num].end_pfn) {
			(*zone_num)++;
			while (*zone_num < (1 << ZONEID_SHIFT) &&
				 !suspend2_nosave_zone_list[*zone_num].end_pfn)
				(*zone_num)++;
			
			if (*zone_num == (1 << ZONEID_SHIFT))
				return suspend2_nosave_max_pfn + 1;
			counter= suspend2_nosave_zone_list[*zone_num].start_pfn;
			reset_ul_ptr = 1;
		} else
			if (!(counter & BIT_NUM_MASK))
				reset_ul_ptr = 1;

		pagebit = PAGEBIT(counter);

		if (reset_ul_ptr) {
			reset_ul_ptr = 0;
			ul_ptr = PAGE_UL_PTR(bitmap, *zone_num, (counter -
			    suspend2_nosave_zone_list[*zone_num].start_pfn));
		}
		
		if (!(*ul_ptr & ~((1 << pagebit) - 1))) {
			counter += BITS_PER_LONG - pagebit - 1;
			continue;
		}
	} while(!test_bit(pagebit, ul_ptr));

	return counter;
}

/**
 * copyback_high: Restore highmem pages.
 *
 * Iterate through the source and destination bitmaps, restoring
 * highmem pages that were atomically copied.
 **/

static void copyback_high(void)
{
	unsigned long *origpage;
	unsigned long *copypage;

	origoffset = __suspend_get_next_bit_on(suspend2_nosave_origmap,
				&o_zone_num, suspend2_nosave_max_pfn + 1);
	copyoffset = __suspend_get_next_bit_on(suspend2_nosave_copymap,
				&c_zone_num, suspend2_nosave_max_pfn + 1);

	while (o_zone_num < (1 << ZONEID_SHIFT)) {
		if (suspend2_nosave_zone_list[o_zone_num].is_highmem) {
			origpage = (unsigned long *) kmap_atomic(
					pfn_to_page(origoffset), KM_USER1);
			copypage = (unsigned long *)
					__va(copyoffset << PAGE_SHIFT);

			memcpy(origpage, copypage, PAGE_SIZE);

			kunmap_atomic(origpage, KM_USER1);
		}
		
		origoffset = __suspend_get_next_bit_on(
				suspend2_nosave_origmap, &o_zone_num,
				origoffset);
		copyoffset = __suspend_get_next_bit_on(
				suspend2_nosave_copymap, &c_zone_num,
				copyoffset);
	}
}
#else /* CONFIG_HIGHMEM */
void copyback_high(void) { }
int suspend_init_nosave_highmem_info(void) { return 0; }
void free_nosave_highmem_info(void) { }
#endif

/**
 * free_pbe_list: Free page backup entries used by the atomic copy code.
 *
 * Normally, this function isn't used. If, however, we need to abort before
 * doing the atomic copy, we use this to free the pbes previously allocated.
 **/
static void free_pbe_list(void)
{
	struct pbe *free_pbe;

	free_pbe = restore_pblist;

	while (restore_pblist) {
		if ((((unsigned long) restore_pblist) & PAGE_MASK) != 
				(unsigned long) free_pbe) {
			free_page((unsigned long) free_pbe);
			free_pbe = restore_pblist;
		}
		restore_pblist = restore_pblist->next;
	}

	if (free_pbe)
		free_page((unsigned long) free_pbe);
}

/*
 * prepare_suspend2_pbe_list
 *
 * Prepare pageset2 pages for doing the atomic copy. If necessary,
 * we allocate extra pages.
 */

static int prepare_suspend2_pbe_list(void)
{
	unsigned long orig_pfn, copy_pfn;
	struct pbe *this_pbe = NULL, *last_pbe = NULL;

	orig_pfn = copy_pfn = (max_pfn + 1);

	restore_pblist = NULL;

	do {
		/* Need new page? */
		if (!this_pbe ||
		    ((((unsigned long) this_pbe) & (PAGE_SIZE - 1)) 
		     + 2 * sizeof(struct pbe)) > PAGE_SIZE) {
			/* Get the next page for pbes */
			this_pbe =
			  (struct pbe *) __suspend_get_nonconflicting_page();
			if (!this_pbe)
				return -ENOMEM;
		} else
			this_pbe++;

		/* Another page to copy? */
		do {
			orig_pfn = get_next_bit_on(pageset1_map, orig_pfn);
			copy_pfn = get_next_bit_on(pageset1_copy_map, copy_pfn);
			if (orig_pfn > max_pfn) {
				if (copy_pfn <= max_pfn) {
					printk("Suspend2: Ran out of original "
							"addresses.");
					return 1;
				}
				return 0;
			}
			if (copy_pfn > max_pfn) {
				printk("Suspend2: Ran out of copy "
						"addresses.\n");
				return 1;
			}
		} while (PageHighMem(pfn_to_page(orig_pfn)));
		
		/* Link the new pbe. */
		if (!last_pbe)
			restore_pblist = this_pbe;
		else
			last_pbe->next = this_pbe;

		/* And fill */
		last_pbe = this_pbe;
		this_pbe->orig_address = page_address(pfn_to_page(orig_pfn));
		this_pbe->address = page_address(pfn_to_page(copy_pfn));
		/* get_nonconflicting_page doesn't get zeroed pages */
		this_pbe->next = NULL;

	} while (1);

	return 0;
}

/**
 * copyback_post: Post atomic-restore actions.
 *
 * After doing the atomic restore, we have a few more things to do:
 * 1) We want to retain some values across the restore, so we now copy
 * these from the nosave variables to the normal ones.
 * 2) Set the status flags.
 * 3) Resume devices.
 * 4) Get userui to redraw.
 * 5) Reread the page cache.
 **/

void copyback_post(void)
{
	int loop;

	suspend_action = suspend2_nosave_state1;
	suspend_debug_state = suspend2_nosave_state2;
	console_loglevel = suspend2_nosave_state3;

	for (loop = 0; loop < 4; loop++)
		suspend_io_time[loop/2][loop%2] =
			suspend2_nosave_io_speed[loop/2][loop%2];

	set_suspend_state(SUSPEND_NOW_RESUMING);
	set_suspend_state(SUSPEND_PAGESET2_NOT_LOADED);

	if (pm_ops && pm_ops->finish && suspend_powerdown_method > 3)
		pm_ops->finish(suspend_powerdown_method);

	if (suspend_activate_storage(1))
		panic("Failed to reactivate our storage.");

	userui_redraw();

	suspend_cond_pause(1, "About to reload secondary pagedir.");

	if (read_pageset2(0))
		panic("Unable to successfully reread the page cache.");

	clear_suspend_state(SUSPEND_PAGESET2_NOT_LOADED);
}

/**
 * suspend_copy_pageset1: Do the atomic copy of pageset1.
 *
 * Make the atomic copy of pageset1. We can't use copy_page (as we once did)
 * because we can't be sure what side effects it has. On my old Duron, with
 * 3DNOW, kernel_fpu_begin increments preempt count, making our preempt
 * count at resume time 4 instead of 3.
 * 
 * We don't want to call kmap_atomic unconditionally because it has the side
 * effect of incrementing the preempt count, which will leave it one too high
 * post resume (the page containing the preempt count will be copied after
 * its incremented. This is essentially the same problem.
 **/

void suspend_copy_pageset1(void)
{
	int i;
	unsigned long source_index, dest_index;

	source_index = get_next_bit_on(pageset1_map, max_pfn + 1);
	dest_index = get_next_bit_on(pageset1_copy_map, max_pfn + 1);

	for (i = 0; i < pagedir1.size; i++) {
		unsigned long *origvirt, *copyvirt;
		struct page *origpage;
		int loop = (PAGE_SIZE / sizeof(unsigned long)) - 1;

		origpage = pfn_to_page(source_index);
		
	       	if (PageHighMem(origpage))
			origvirt = kmap_atomic(origpage, KM_USER0);
		else
			origvirt = page_address(origpage);

		copyvirt =
			(unsigned long *) page_address(pfn_to_page(dest_index));

		while (loop >= 0) {
			*(copyvirt + loop) = *(origvirt + loop);
			loop--;
		}
		
		if (PageHighMem(origpage))
			kunmap_atomic(origvirt, KM_USER0);
		
		source_index = get_next_bit_on(pageset1_map, source_index);
		dest_index = get_next_bit_on(pageset1_copy_map, dest_index);
	}
}

/**
 * __suspend_post_context_save: Steps after saving the cpu context.
 *
 * Steps taken after saving the CPU state to make the actual
 * atomic copy.
 *
 * Called from swsusp_save in snapshot.c via suspend_post_context_save.
 **/

int __suspend_post_context_save(void)
{
	int old_ps1_size = pagedir1.size;
	
	calculate_check_checksums(1);

	free_checksum_pages();

	suspend_recalculate_image_contents(1);

	extra_pd1_pages_used = pagedir1.size - old_ps1_size;

	if (extra_pd1_pages_used > extra_pd1_pages_allowance) {
		printk("Pageset1 has grown by %ld pages. "
			"extra_pages_allowance is currently only %d.\n",
			pagedir1.size - old_ps1_size,
			extra_pd1_pages_allowance);
		set_result_state(SUSPEND_ABORTED);
		set_result_state(SUSPEND_EXTRA_PAGES_ALLOW_TOO_SMALL);
		return -1;
	}

	if (!test_action_state(SUSPEND_TEST_FILTER_SPEED) &&
	    !test_action_state(SUSPEND_TEST_BIO))
		suspend_copy_pageset1();

	return 0;
}

/**
 * suspend2_suspend: High level code for doing the atomic copy.
 *
 * High-level code which prepares to do the atomic copy. Loosely based
 * on the swsusp version, but with the following twists:
 * - We set suspend2_running so the swsusp code uses our code paths.
 * - We give better feedback regarding what goes wrong if there is a problem.
 * - We use an extra function to call the assembly, just in case this code
 *   is in a module (return address).
 **/

int suspend2_suspend(void)
{
	int error;

	suspend2_running = 1; /* For the swsusp code we use :< */

	if (test_action_state(SUSPEND_PM_PREPARE_CONSOLE))
		pm_prepare_console();

	if ((error = arch_prepare_suspend()))
		goto err_out;

	local_irq_disable();

	/* At this point, device_suspend() has been called, but *not*
	 * device_power_down(). We *must* device_power_down() now.
	 * Otherwise, drivers for some devices (e.g. interrupt controllers)
	 * become desynchronized with the actual state of the hardware
	 * at resume time, and evil weirdness ensues.
	 */
	if ((error = device_power_down(PMSG_FREEZE))) {
		set_result_state(SUSPEND_DEVICE_REFUSED);
		set_result_state(SUSPEND_ABORTED);
		printk(KERN_ERR "Some devices failed to power down, aborting suspend\n");
		goto enable_irqs;
	}

	save_processor_state();
	if ((error = swsusp_arch_suspend()))
		printk(KERN_ERR "Error %d suspending\n", error);
	/* Restore control flow appears here */
	restore_processor_state();

	if (!suspend2_in_suspend)
		copyback_high();

	device_power_up();
enable_irqs:
	local_irq_enable();
	if (test_action_state(SUSPEND_PM_PREPARE_CONSOLE))
		pm_restore_console();
err_out:
	suspend2_running = 0;
	return error;
}

/**
 * suspend_atomic_restore: Prepare to do the atomic restore.
 *
 * Get ready to do the atomic restore. This part gets us into the same
 * state we are in prior to do calling do_suspend2_lowlevel while
 * suspending: hot-unplugging secondary cpus and freeze processes,
 * before starting the thread that will do the restore.
 **/

int suspend_atomic_restore(void)
{
	int error, loop;

	suspend2_running = 1;

	suspend_prepare_status(DONT_CLEAR_BAR,	"Atomic restore preparation");
	if (prepare_suspend2_pbe_list())
		goto free_pbe_list;

	if (test_action_state(SUSPEND_PM_PREPARE_CONSOLE))
		pm_prepare_console();

	disable_nonboot_cpus();

	if ((error = freeze_processes())) {
		printk("Some processes failed to suspend\n");
		goto thaw_processes;
	}

	if ((error = device_suspend(PMSG_FREEZE))) {
		printk("Some devices failed to suspend\n");
		goto device_resume;
	}

	if (suspend_init_nosave_highmem_info())
		goto free_nosave_highmem_info;

	suspend2_nosave_state1 = suspend_action;
	suspend2_nosave_state2 = suspend_debug_state;
	suspend2_nosave_state3 = console_loglevel;
	
	for (loop = 0; loop < 4; loop++)
		suspend2_nosave_io_speed[loop/2][loop%2] =
			suspend_io_time[loop/2][loop%2];
	memcpy(suspend2_nosave_commandline, saved_command_line, COMMAND_LINE_SIZE);

	mb();

	local_irq_disable();
	if (device_power_down(PMSG_FREEZE)) {
		printk(KERN_ERR "Some devices failed to power down. Very bad.\n");
		goto device_power_up;
	}

	/* We'll ignore saved state, but this gets preempt count (etc) right */
	save_processor_state();
	error = swsusp_arch_resume();
	/* 
	 * Code below is only ever reached in case of failure. Otherwise
	 * execution continues at place where swsusp_arch_suspend was called.
	 *
	 * We don't know whether it's safe to continue (this shouldn't happen),
	 * so lets err on the side of caution.
         */
	BUG();

device_power_up:
	device_power_up();
free_nosave_highmem_info:
	free_nosave_highmem_info();
device_resume:
	device_resume();
thaw_processes:
	thaw_processes();
	enable_nonboot_cpus();
	if (test_action_state(SUSPEND_PM_PREPARE_CONSOLE))
		pm_restore_console();
free_pbe_list:
	free_pbe_list();
	return 1;
}
